Package: Require
Type: Package
Title: Installing and Loading R Packages for Reproducible Workflows
Description: A single key function, 'Require' that makes rerun-tolerant
    versions of 'install.packages' and 'require' for CRAN packages, packages
    no longer on CRAN (i.e., archived), specific versions of packages, 
    and GitHub packages. This approach is developed to create reproducible 
    workflows that are flexible and fast enough to use while in development stages,
    while able to build snapshots once a stable package collection is found. 
    As with other functions in a reproducible workflow, this package 
    emphasizes functions that return the same result whether it is 
    the first or subsequent times running the function, with subsequent times being
    sufficiently fast that they can be run every time without undue waiting burden on 
    the user or developer.
URL: https://Require.predictiveecology.org,
        https://github.com/PredictiveEcology/Require
Date: 2026-09-03
Version: 2.1.1
Authors@R: c(
    person(given = "Eliot J B",
           family = "McIntire",
           role = c("aut", "cre"),
           email = "eliot.mcintire@canada.ca",
           comment = c(ORCID = "0000-0002-6914-8316")),
      person(given = "Alex M",
             family = "Chubaty",
             role = "ctb",
             email = "achubaty@for-cast.ca",
             comment = c(ORCID = "0000-0001-7146-8135")),
    person(paste("His Majesty the King in Right of Canada,",
                 "as represented by the Minister of Natural Resources Canada"),
           role = "cph")
    )
Depends: R (>= 4.1)
Imports: callr, data.table (>= 1.10.4), methods, pak, processx, sys,
        tools, utils
Suggests: cli, covr, curl, diffobj, fpCompare, gitcreds, httr,
        parallel, pkgcache, pkgload, rematch2, rmarkdown, knitr, rlang,
        roxygen2, rprojroot, testthat (>= 3.0.0), tibble, waldo, withr
Encoding: UTF-8
Language: en-CA
License: GPL-3
VignetteBuilder: knitr, rmarkdown
BugReports: https://github.com/PredictiveEcology/Require/issues
ByteCompile: yes
Config/roxygen2/version: 8.1.0
Collate: 'CRAN.R' 'Require-helpers.R' 'Require-package.R' 'messages.R'
        'Require2.R' 'RequireOptions.R' 'envs.R' 'extract.R'
        'helpers.R' 'pak.R' 'pkgDep.R' 'pkgDep3.R' 'pkgSnapshot.R'
        'setLibPaths.R' 'setup.R' 'zzz.R'
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-09-03 18:12:24 UTC; emcintir
Author: Eliot J B McIntire [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6914-8316>),
  Alex M Chubaty [ctb] (ORCID: <https://orcid.org/0000-0001-7146-8135>),
  His Majesty the King in Right of Canada, as represented by the Minister
    of Natural Resources Canada [cph]
Maintainer: Eliot J B McIntire <eliot.mcintire@canada.ca>
Repository: CRAN
Date/Publication: 2026-09-04 05:10:59 UTC
