# Generated by roxygen2: do not edit by hand

export(.downloadFileMasterMainAuth)
export(GETWauthThenNonAuth)
export(Install)
export(Require)
export(RequireOptions)
export(cacheClearPackages)
export(cacheDefaultDir)
export(cacheDir)
export(cacheGetOptionCachePkgDir)
export(cachePkgDir)
export(cachePurge)
export(checkPath)
export(clearRequirePackageCache)
export(compareVersion2)
export(detachAll)
export(dlArchiveVersionsAvailable)
export(dlGitHubDESCRIPTION)
export(extractInequality)
export(extractPkgGitHub)
export(extractPkgName)
export(extractVersionNumber)
export(getCRANrepos)
export(getGitCredsToken)
export(getRequireOptions)
export(invertList)
export(messageDF)
export(messageVerbose)
export(modifyList2)
export(modifyList3)
export(normPath)
export(paddedFloatToChar)
export(parseGitHub)
export(pkgDep)
export(pkgDep2)
export(pkgDepIfDepRemoved)
export(pkgDepTopoSort)
export(pkgSnapshot)
export(pkgSnapshot2)
export(purgeCache)
export(setLibPaths)
export(setLinuxBinaryRepo)
export(setdiffNamed)
export(setup)
export(setupOff)
export(sourcePkgs)
export(tempdir2)
export(tempfile2)
export(trimRedundancies)
export(trimVersionNumber)
export(updatePackages)
exportMethods(checkPath)
exportMethods(normPath)
import(methods)
importFrom(callr,r)
importFrom(data.table,
  ":=",
  .I,
  .SD,
  as.data.table,
  copy,
  data.table,
  fwrite,
  is.data.table,
  rbindlist,
  rleid,
  set,
  setDT,
  setnames,
  setorderv
)
importFrom(processx,run)
importFrom(stats,
  na.omit,
  setNames
)
importFrom(sys,
  exec_background,
  exec_wait
)
importFrom(tools,R_user_dir)
importFrom(utils,
  assignInMyNamespace,
  available.packages,
  capture.output,
  chooseCRANmirror,
  compareVersion,
  contrib.url,
  download.file,
  head,
  install.packages,
  installed.packages,
  modifyList,
  packageDescription,
  packageVersion,
  read.csv,
  remove.packages,
  sessionInfo,
  tail,
  untar,
  unzip,
  write.table
)
