% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/envs.R
\name{envPkgDepDESCFileCreate}
\alias{envPkgDepDESCFileCreate}
\title{3rd level for DESCRIPTIONFile}
\usage{
envPkgDepDESCFileCreate()
}
\description{
3rd level for DESCRIPTIONFile
}
